# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Miscellaneous utilities."""


def truthy(val: str | int | bool) -> bool:
    """Returns True if val is truthy, else False.

    Truthy values include:
        - "true" in any combination of capitalizations
        - "1"
        - 1
        - True
    Args:
        val: Value to check for truthiness.
    """
    return str(val).lower() in ["true", "1"]



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Schema models for project configuration specified in aineko.yml.

These models are used to validate the project configuration specified in
repos that are used in `aineko create`.
"""

from typing import Optional

from pydantic import BaseModel, field_validator

from aineko import __version__


class ProjectConfig(BaseModel):
    """Project configuration (Schema for aineko.yml)."""

    aineko_version: str
    project_name: str
    project_slug: Optional[str] = None
    project_description: Optional[str] = None
    pipeline_slug: Optional[str] = None

    @field_validator("aineko_version")
    def version(cls, v: str) -> str:  # pylint: disable=no-self-argument
        """Validates that the aineko version matches."""
        if v != __version__:
            raise ValueError(
                f"Project config `aineko.yml` requires version {v}, "
                f"but current version is {__version__}. Use "
                f"`pip install aineko=={v}` to install the correct version."
            )
        return v



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A series of cookiecutter post-codegen hooks to validate user input.

These code will run after files are generated

See: https://cookiecutter.readthedocs.io/en/1.7.2/advanced/hooks.html
"""

import os

import yaml
from github import Auth, Github

from aineko import __version__
from aineko.models.project_config_schema import ProjectConfig


def remove_deploy_file():
    """Remove deploy file if it exists."""
    try:
        deploy_file_path = os.path.join(os.getcwd(), "deploy.yml")
        if os.path.isfile(deploy_file_path):
            os.remove(deploy_file_path)
    except FileNotFoundError:
        pass


def create_github_client() -> Github:
    """Creates a GitHub client.

    Returns:
        GitHub client.
    """
    if os.getenv("GITHUB_TOKEN"):
        return Github(auth=Auth.Token(os.environ["GITHUB_TOKEN"]))
    return Github()


def get_all_repo_contents(full_repo_rev: str) -> list:
    """Recursively gets all contents of a GitHub repo.

    Args:
        full_repo_rev: GitHub repo to clone,
            in the format of <<owner>>/<<repo>>#<<rev>>.

    Returns:
        List of all contents of the repo.
    """
    g = create_github_client()
    repo_rev = full_repo_rev.split("#")
    repo = g.get_repo(repo_rev[0])

    all_contents = []
    contents = repo.get_contents("", ref=repo_rev[1])
    while contents:
        file_content = contents.pop(0)
        if file_content.type == "dir":
            contents.extend(
                repo.get_contents(file_content.path, ref=repo_rev[1])
            )
        else:
            all_contents.append(file_content)
    g.close()
    return all_contents


def get_file_from_repo(full_repo_rev: str, file_path: str) -> str:
    """Gets a file from a GitHub repo.

    Args:
        full_repo_rev: GitHub repo to clone,
            in the format of <<owner>>/<<repo>>#<<rev>>.
        file_path: Path to file to get.

    Returns:
        Contents of file.
    """
    g = create_github_client()
    repo_rev = full_repo_rev.split("#")
    repo = g.get_repo(repo_rev[0])
    g.close()
    return repo.get_contents(file_path, ref=repo_rev[1]).decoded_content


def add_files_from_repo(full_repo_rev: str, project_slug: str):
    """Adds relevant files from the GitHub repo revision.

    Args:
        full_repo_rev: GitHub repo to clone,
            in the format of <<owner>>/<<repo>>#<<rev>>.
        project_slug: Project slug used for the generated project.
    """
    contents = get_all_repo_contents(full_repo_rev)

    # Parse aineko.yml
    project_config_raw = get_file_from_repo(full_repo_rev, "aineko.yml")
    project_config = yaml.safe_load(project_config_raw)
    project_config = ProjectConfig(**project_config)

    # Remove files that will be replaced
    os.remove(os.path.join(project_slug, "nodes.py"))
    os.remove(os.path.join("conf", "pipeline.yml"))

    # Add all files except aineko.yml
    for content in contents:
        if content.name == "aineko.yml":
            continue

        file_path = f"{content.path}"
        try:
            os.remove(file_path)
        except FileNotFoundError:
            pass

        if os.path.dirname(file_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as file:
            file.write(content.decoded_content)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Utils to create a aineko pipeline from existing templates."""

import os

import click
import yaml
from cookiecutter.main import cookiecutter  # type: ignore

from aineko.models.project_config_schema import ProjectConfig
from aineko.templates.first_aineko_pipeline.hooks.utils import (
    get_file_from_repo,
)


@click.command()
@click.option(
    "-d",
    "--deployment-config",
    is_flag=True,
    help="Include deploy.yml that facilitates deployment of pipelines.",
)
@click.option(
    "-o",
    "--output-dir",
    default=".",
    help="Directory to create pipeline in. Defaults to current directory.",
)
@click.option(
    "-n",
    "--no-input",
    is_flag=True,
    help="Do not prompt for parameters and use defaults.",
)
@click.option(
    "-r",
    "--repo",
    help=(
        "GitHub repo to create pipeline from. Link should contain branch or "
        "rev to reference. "
        "(Ex aineko-dev/dream-catcher#my-branch) "
        "Refer to CLI docs for more info on repo structure."
    ),
)
def create(
    deployment_config: bool, output_dir: str, no_input: bool, repo: str
) -> None:
    """Creates boilerplate code and config required for an Aineko pipeline.\f

    If repo is set, attempts to clone and use the files from the repo.
    The following files are required:

        - aineko.yml: Contains required data to create an aineko pipeline.
            Requires keys:
            - aineko_version: Version of aineko to use.
            - project_name: Name of the aineko project.
            - project_slug: (optional) slug of project. Can only contain
                alphanumeric characters and underscores. Will be derived from
                project_name if not provided.
            - project_description: (optional) Description of the pipeline.
            - pipeline_slug: (optional) name of pipeline (alphanumeric
                characters, dashes and underscores only).
        - <<project_name>>/nodes.py or <<project_name>>/nodes/*.py: Either a
            file containing all node code or a directory containing multiple
            node code files.
        - conf/*.yml: Directory containing config files for pipelines.

    The following files are optional, and will overwrite the default files:
        - README.md: README file for the repo.
        - pyproject.toml: Project configuration, including poetry requirements
        - deploy.yml: Project deployment configuration.
        - <<project_name>>/tests/*.py: Test directory

    Args:
        deployment_config: If True, include a deploy file when generating, else
        do not include.
        output_dir: Directory to create pipeline in. Defaults to current.
        no_input: If True, do not prompt for parameters and use defaults.
        repo: GitHub repo to create pipeline from. Link should contain branch
            or rev to reference.
    """
    script_directory = os.path.dirname(os.path.abspath(__file__))
    # assumes templates is a folder that is one-level up
    templates_directory = f"{os.path.dirname(script_directory)}/templates"

    extra_context = {
        "_deployment_config": deployment_config,
        "_repo": repo,
    }

    if repo:
        click.echo("Getting project config from repo...")
        no_input = True
        project_config_raw = get_file_from_repo(repo, "aineko.yml")
        project_config = yaml.safe_load(project_config_raw)
        project_config = ProjectConfig(**project_config)

        extra_context["project_name"] = project_config.project_name
        extra_context[
            "project_description"
        ] = project_config.project_description
        if project_config.project_slug:
            extra_context["project_slug"] = project_config.project_slug
        if project_config.pipeline_slug:
            extra_context["pipeline_slug"] = project_config.pipeline_slug

    cookiecutter(
        f"{templates_directory}/first_aineko_pipeline",
        extra_context=extra_context,
        output_dir=output_dir,
        no_input=no_input,
    )

    click.echo("Successfully created pipeline.")



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A wrapper around API calls to the Aineko Dream API."""

import click
import requests


@click.command()
@click.argument("prompt")
@click.argument("api-key")
@click.option(
    "-u",
    "--url",
    default="https://dream.ap-6d356ccd-bfba-4110-994a-fe164ab8bf77.aineko.app",
    help="API url to use for the Aineko Dream API.",
)
def dream(prompt: str, api_key: str, url: str) -> None:
    """Main function to generate an aineko project using Aineko Dream.\f

    Args:
        prompt: Prompt to generate a project from.
        api_key: API key to use for the Aineko Dream API.
        url: URL to use for the Aineko Dream API.
    """
    try:
        click.echo("Generating project from prompt...")

        r = requests.get(
            url=f"{url}/create/",
            params={"prompt": prompt},
            headers={"x-api-key": api_key},
            timeout=300,
        )
        click.echo(f"Received response code: {r.status_code}")
        if r.status_code == 200:
            request_id = r.json()["request_id"]
            click.echo(
                "Successfully published project to: "
                f"https://github.com/Convex-Labs/dream-catcher/tree/{request_id}. \n\n"  # pylint: disable=line-too-long
                "To create a new project from this dream, run: \n\n"
                f"aineko create --repo Convex-Labs/dream-catcher#{request_id}"
            )
        else:
            click.echo(r.text)

    except Exception as e:
        raise click.ClickException(
            f"Error connecting to Aineko Dream API at {url}: {e}"
        )



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A series of cookiecutter post-codegen hooks to validate user input.

These code will run after files are generated

See: https://cookiecutter.readthedocs.io/en/1.7.2/advanced/hooks.html
"""

from aineko.templates.first_aineko_pipeline.hooks.utils import (
    add_files_from_repo,
    remove_deploy_file,
)
from aineko.utils.misc import truthy

with_deploy = "{{ cookiecutter._deployment_config }}"
repo = "{{ cookiecutter._repo }}"
project_slug = "{{ cookiecutter.project_slug }}"

if not truthy(with_deploy):
    remove_deploy_file()

if repo != "None":
    add_files_from_repo(repo, project_slug)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Extra module for running a FastAPI server.

This module contains the Consumers and Producers class that give
access to the node's consumers and producers from within the FastAPI
app. It also contains the FastAPI node class that runs the uvicorn
server.

We recommend no more than 1 FastAPI node per pipeline since the Consumer
and Producer objects are namespaced at the pipeline level. If you must
have multiple FastAPI nodes, we recommend using different datasets to avoid
namespace collisions.
"""

from typing import Optional

import uvicorn

from aineko import AbstractNode, DatasetConsumer, DatasetProducer


class Consumers(dict):
    """Class to contain consumers."""

    def __setitem__(
        self, key: str | int | tuple, value: DatasetConsumer
    ) -> None:
        """Checks that item is of type DatasetConsumer before setting.

        Args:
            key: Name of the dataset
            value: DatasetConsumer object to be stored

        Raises:
            ValueError: If value is not of type DatasetConsumer
        """
        if not isinstance(value, DatasetConsumer):
            raise ValueError(
                f"Value must be of type DatasetConsumer, not {type(value)}"
            )
        super().__setitem__(key, value)


class Producers(dict):
    """Class to contain producers."""

    def __setitem__(
        self, key: str | int | tuple, value: DatasetProducer
    ) -> None:
        """Checks that item is of type DatasetProducer before setting.

        Args:
            key: Name of the dataset
            value: DatasetProducer object to be stored

        Raises:
            ValueError: If value is not of type DatasetProducer
        """
        if not isinstance(value, DatasetProducer):
            raise ValueError(
                f"Value must be of type DatasetProducer, not {type(value)}"
            )
        super().__setitem__(key, value)


consumers = Consumers()
producers = Producers()


# pylint: disable=anomalous-backslash-in-string
class FastAPI(AbstractNode):
    """Node for creating a FastAPI app with a gunicorn server.

    `node_params` should contain the following keys:

        app: path to FastAPI app
        port (optional): port to run the server on. Defaults to 8000.
        log_level (optional): log level to log messages from the uvicorn server.
            Defaults to "info".

    To access the consumers and producers from your FastAPI app, import the
    `consumers` and `producers` variables from `aineko.extras.fastapi`. Use
    them as you would use `self.consumers` and `self.producers` in a regular
    node.

    We recommend no more than 1 FastAPI node per pipeline since the Consumer
    and Producer objects are namespaced at the pipeline level.

    Example usage in pipeline.yml:
    ```yaml title="pipeline.yml"
    pipeline:
      nodes:
        fastapi:
          class: aineko.extras.FastAPI
          inputs:
            - test_sequence
          node_params:
            app: my_awesome_pipeline.fastapi:app
            port: 8000
    ```
    where the app points to a FastAPI app. See
    [FastAPI documentation](https://fastapi.tiangolo.com/){:target="\_blank"}
    on how to create a FastAPI app.

    Example usage in FastAPI app:
    ```python title="fastapi.py"
    from aineko.extras.fastapi import consumers, producers

    @app.get("/query")
    async def query():
        msg = consumers["test_sequence"].next()
        return msg
    ```
    """

    def _pre_loop_hook(self, params: Optional[dict] = None) -> None:
        """Initialize node state. Set env variables for Fast API app."""
        for key, value in self.consumers.items():
            consumers[key] = value
        for key, value in self.producers.items():
            producers[key] = value

    def _execute(self, params: dict) -> None:
        """Start the API server."""
        config = uvicorn.Config(
            app=params.get("app"),  # type: ignore
            port=params.get("port", 8000),
            log_level=params.get("log_level", "info"),
            host="0.0.0.0",
        )
        server = uvicorn.Server(config)
        server.run()
